import asyncio
import json
import os
import faster_than_requests as requests
from web3 import Web3, WebsocketProvider, Account
from web3.gas_strategies.time_based import slow_gas_price_strategy,medium_gas_price_strategy, fast_gas_price_strategy
from xchainpy_ethereum.models.asset import Asset
from xchainpy_client import interface
from xchainpy_crypto import crypto


class IEthereumClient:
    def is_connected(self):
        pass

    async def get_abi(self, contract_address):
        pass

    async def get_contract(self, contract_address, erc20=True):
        pass

    async def read_contract(self, contract_address, func_to_call, *args, erc20=True):
        pass

    async def write_contract(self, contract_address, func_to_call, *args, erc20=True, gas_limit=1000000, gas_price=None, nonce=None):
        pass

    def set_gas_strategy(self, gas_strategy):
        pass

    async def transfer(self, asset: Asset, amount, recipient, gas_limit=1000000, gas_price=None):
        pass

    def get_transaction_data(self, tx_id):
        pass

    def get_transaction_receipt(self, tx_id):
        pass

    async def get_balance(self, asset: Asset=None, address=None):
        pass


class Client(interface.IXChainClient, IEthereumClient):
    script_dir = os.path.dirname(__file__)
    with open(os.path.join(script_dir, "resources/ERC20"), 'r') as f:
        erc20_abi = json.loads(f.read())["abi"]
    network = network_type = ether_api = ''
    gas_strategy = "medium"
    gas_price = None
    w3 = account = None

    def __init__(self,  phrase: str, network: str, network_type: str = "ropsten", ether_api: str = None):
        """Constructor

        Client has to be initialised with mnemonic phrase, network (infura_api token), network_type ("mainnet" or "ropsten"),
        ether_api: etherscan api token.
        It will throw an error if an invalid phrase or network has been passed.

        Args:
            phrase: phrase of wallet (mnemonic) will be set to the Class
            network: infura websocket api endpoint of the selected network_type
            network_type: network type can either be `mainnet` or 'ropsten'
            ether_api: etherscan API token, used for downloading non ERC20 contract ABI

        Returns:
            void

        """
        if network_type != "ropsten" and network_type != "mainnet":
            raise Exception('Network type has to be ropsten or mainnet')
        self.ether_api = ether_api
        self.network_type = network_type
        os.makedirs(os.path.join(self.script_dir, f'resources/{network_type}'), exist_ok=True)
        self.set_network(network)
        self.set_phrase(phrase)

    def purge_client(self):
        """Purge Client

        Delete Account

        Returns:
            void

        """
        self.w3 = self.account = None

    def is_connected(self):
        """Check Web3 connectivity

        Returns:
            bool

        """
        return self.w3.isConnected()

    def set_network(self, network: str):
        """Set/update the current network

        It will throw an error if an invalid phrase or network has been passed.

        Args:
            network: infura websocket api endpoint of the selected network_type

        Returns:
            void

        Raises:
            Exception: "Network must be provided". -> Thrown if network has not been set before.

        """
        self.network = network
        if self.network_type not in network:
            raise Exception("invalid network type")
        self.w3 = Web3(WebsocketProvider(network))
        if not self.is_connected():
            raise Exception("Infura API error")

    def get_network(self):
        """Get the current network

        Returns:
            infura websocket api

        """
        return self.network

    def validate_address(self, address: str):
        """Check address validity

        Args:
            address: ethereum address

        Returns:
            bool

        """
        return self.w3.isAddress(address)

    def get_address(self):
        """Get current wallet address

        Returns:
            current wallet address

        """
        return self.account.address

    def set_phrase(self, phrase: str):
        """Set/Update a new phrase

        Args:
            phrase: A new phrase

        Returns:
            The address of the given phrase

        Raises:
            'Invalid Phrase' if the given phrase is invalid

        """
        if not crypto.validate_phrase(phrase):
            raise Exception("invalid phrase")
        Account.enable_unaudited_hdwallet_features()
        self.account = self.w3.eth.account.from_mnemonic(mnemonic=phrase)
        return self.get_address()

    async def get_abi(self, contract_address):
        """Get abi description of a non ERC-20 contract

        Args:
            contract_address: contract address

        Returns:
            abi description[json]

        """
        path = os.path.join(self.script_dir, f'resources/{self.network_type}/{contract_address}')
        if os.path.exists(path):
            with open(path, 'r') as f:
                return json.loads(f.read())
        else:
            resource_path = os.path.join(self.script_dir, f'resources/{self.network_type}')
            os.makedirs(resource_path, exist_ok=True)
            if not self.ether_api:
                raise Exception("undefined ether api token")
            if self.network_type == 'mainnet':
                url = f'https://api.etherscan.io/api?module=contract&action=getabi&address={contract_address}&apikey={self.ether_api}'
            else:
                url = f'https://api-ropsten.etherscan.io/api?module=contract&action=getabi&address={contract_address}&apikey={self.ether_api}'
            r = json.loads(requests.get2json(url))
            if r["status"] != '1':
                raise Exception("error getting abi file")
            with open(path, 'w+') as o:
                o.write(r["result"])
            return json.loads(r["result"])

    async def get_contract(self, contract_address, erc20=True):
        """Get Contract object of given address

        Args:
            contract_address: ethereum contract address
            erc20: True if contract = ERC-20, False otherwise

        Returns:
            web3 contract object

        """
        if not erc20:
            contract = await self.get_abi(contract_address)
        else:
            contract = self.erc20_abi
        return self.w3.eth.contract(abi=contract, address=contract_address)

    async def get_balance(self, asset=None, address=None):
        """Get the balance of a erc-20 token

        Args:
            address: By default, it will return the balance of the current wallet. (optional)
            contract_address: If not set, it will return ethereum balance. (optional)

        Returns:
            Balance of the address

        """
        if not asset:
            if address:
                return self.w3.fromWei(self.w3.eth.get_balance(address), 'ether')
            return self.w3.fromWei(self.w3.eth.get_balance(self.get_address()), 'ether')
        else:
            assert asset.chain == 'ETH'
            token_contract = await self.get_contract(asset.ticker)
            decimal = token_contract.functions.decimals().call()
            if address:
                return token_contract.functions.balanceOf(address).call() / 10 ** decimal
            return token_contract.functions.balanceOf(self.get_address()).call() / 10 ** decimal

    def set_gas_strategy(self, gas_strategy) -> None:
        """Set Gas fee calculation parameter

        fast: transaction mined within 60 seconds
        medium: transaction mined within 5 minutes
        slow: transaction mined within 1 hour

        Args:
            gas_strategy: ['fast', 'medium', 'slow']

        Returns:
            void
        """
        if gas_strategy == "fast":
            self.w3.eth.setGasPriceStrategy(fast_gas_price_strategy)
        elif gas_strategy == 'medium':
            self.w3.eth.setGasPriceStrategy(medium_gas_price_strategy)
        elif gas_strategy == 'slow':
            self.w3.eth.setGasPriceStrategy(slow_gas_price_strategy)
        else:
            raise Exception("invalid gas strategy")
        self.gas_price = self.w3.eth.generateGasPrice()

    def get_fees(self):
        """Return Gas price using gas_strategy

        Returns:
            gas price in Wei

        """
        return self.gas_price

    async def transfer(self, asset: Asset, amount, recipient, gas_limit=1000000, gas_price=None):
        """Transfer ERC20 token with previous configured gas_price

        Args:
            recipient: recipient address
            amount: amount in ether or in alt coin
            gas_limit: gas limit using gas price
            gas_price: gas price in wei
            contract_address: for assets other than ether

        Returns:
            tx_hash(str)

        """
        if not gas_price:
            gas_price = self.get_fees()
            if not gas_price:
                raise Exception("gas_price not set")
        nonce = self.w3.eth.getTransactionCount(self.get_address())
        if asset.symbol == 'ETH':
            tx = {
                'nonce': nonce,
                'to': recipient,
                'value': self.w3.toWei(amount, 'ether'),
                'gas': gas_limit,
                'gasPrice': gas_price,
            }
            signed_tx = self.account.sign_transaction(tx)
            tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
            return tx_hash
        else:
            tx = {
                'nonce': nonce,
                'gas': gas_limit,
                'gasPrice': gas_price,
            }
            token_contract = await self.get_contract(contract_address=asset.ticker)
            decimal = token_contract.functions.decimals().call()
            raw_tx = token_contract.functions.transfer(recipient, amount*10**decimal).buildTransaction(tx)
            signed_tx = self.account.sign_transaction(raw_tx)
            tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
            receipt = self.w3.eth.waitForTransactionReceipt(tx_hash)
            return receipt

    async def read_contract(self, contract_address, func_to_call, *args, erc20=True):
        contract = await self.get_contract(contract_address=contract_address, erc20=erc20)
        return contract.functions[func_to_call](*args).call()

    async def write_contract(self, contract_address, func_to_call, *args, erc20=True, gas_limit=1000000, gas_price=None, nonce=None, eth_to_be_sent=0):
        """Write to any contract with any argument, specify whether it's ERC20

        Args:
            contract_address: contract address to interact with
            func_to_call: name of contract function to call
            erc20: True if contract = ERC-20, False otherwise
            gas_limit: 1000000 by default
            gas_price: gas price
            **kwargs: arguments for func_to_call
            nonce: provide nonce for faster execution
            eth_to_be_sent: in case ethereum needed to be sent

        Returns:

        """
        if not nonce:
            nonce = self.w3.eth.getTransactionCount(self.get_address())
        if not gas_price:
            gas_price = self.gas_price
        if not gas_price:
            raise Exception("provide gas price or call set_gas_strategy()")
        if eth_to_be_sent != 0:
            tx = {
                'nonce': nonce,
                'value': self.w3.toWei(eth_to_be_sent, 'ether'),
                'gas': gas_limit,
                'gasPrice': gas_price,
            }
        else:
            tx = {
                'nonce': nonce,
                'gas': gas_limit,
                'gasPrice': gas_price,
            }
        smart_contract = await self.get_contract(contract_address=contract_address, erc20=erc20)
        contract_func = smart_contract.functions[func_to_call]
        raw_tx = contract_func(*args).buildTransaction(tx)
        signed_tx = self.account.sign_transaction(raw_tx)
        tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash, timeout=600, )
        return receipt

    def get_transaction_data(self, tx_id):
        """
        Args:
            tx_id:

        Returns:
        AttributeDict({
            'blockHash': '0x4e3a3754410177e6937ef1f84bba68ea139e8d1a2258c5f85db9f1cd715a1bdd',
            'blockNumber': 46147,
            'from': '0xA1E4380A3B1f749673E270229993eE55F35663b4',
            'gas': 21000,
            'gasPrice': 50000000000000,
            'hash': '0x5c504ed432cb51138bcf09aa5e8a410dd4a1e204ef84bfed1be16dfba1b22060',
            'input': '0x',
            'nonce': 0,
            'to': '0x5DF9B87991262F6BA471F09758CDE1c0FC1De734',
            'transactionIndex': 0,
            'value': 31337,
        })
        """
        return self.w3.eth.get_transaction(tx_id)

    def get_transaction_receipt(self, tx_id):
        return self.w3.eth.get_transaction_receipt(tx_id)



from mnemonic import Mnemonic


def generate_mnemonic(language, filename):
    """Generate test_suite using user-specified language and save to local file

    :param language: language of test_suite words
    :type language: str
    :param filename: filename to save test_suite
    :type filename: str
    :returns: test_suite
    """
    words = Mnemonic(language).generate(strength=256)
    output = open(filename, "w+")
    output.write(words)
    output.close()
    return words


generate_mnemonic("english", "resources/mainnet/mnemonic")


import pytest
from xchainpy_ethereum.client import Client
import json, os


class TestClient:
    prefix = "../xchainpy_ethereum/resources/mainnet/"
    mnemonic = open(prefix + "mnemonic", 'r').readline()
    sushi_abi = json.loads(open(prefix + "sushi_abi", 'r').readline())
    token_abi = json.loads(open(prefix + "token_abi", 'r').readline())
    network = open(prefix + "network", 'r').readline()
    eth_api = open("../xchainpy_ethereum/resources/ether_api", 'r').readline()
    sushi_contract_address = "0xcc39592f5cB193a70f262aA301f54DB1d600e6Da"
    thor_token_address = "0x3155BA85D5F96b2d030a4966AF206230e46849cb"
    test_address = "0xB1aA026725cD51700E674E240D971785F95429FD"
    test_hash = "0x1ea72e2c1721757fd431306cb124e2b75943c5cd20fee5e3e1f79e028c00e99e"
    test_rune_transfer_hash = "0xc8b3e4c68c9d7022290d4203ff867dcef157efb8927ad29e718f97ac5dddaab2"

    @pytest.fixture
    def test_init(self):
        self.client = Client(phrase=self.mnemonic, network=self.network, network_type="mainnet",
                             ether_api=self.eth_api)
        yield
        self.client.purge_client()

    def test_invalid_network_type(self):
        with pytest.raises(Exception) as err:
            assert Client(phrase=self.mnemonic, network=self.network, network_type="invalid")
        assert str(err.value) == "Network type has to be ropsten or mainnet"
        with pytest.raises(Exception) as err:
            assert Client(phrase=self.mnemonic, network=self.network, network_type="ropsten")
        assert str(err.value) == "invalid network type"

    def test_invalid_init_phrase(self):
        with pytest.raises(Exception) as err:
            assert Client(phrase="invalid", network=self.network, network_type="mainnet")
        assert str(err.value) == "invalid phrase"

    def test_invalid_set_phrase(self, test_init):
        with pytest.raises(Exception) as err:
            assert self.client.set_phrase(phrase="invalid")
        assert str(err.value) == "invalid phrase"

    def test_is_connected(self, test_init):
        assert self.client.is_connected()

    def test_invalid_init_network(self):
        with pytest.raises(Exception) as err:
            assert Client(phrase=self.mnemonic, network="wss://mainnet.infura.io/ws/v3/invalid")

    def test_invalid_set_network(self, test_init):
        with pytest.raises(Exception) as err:
            assert self.client.set_network(network="wss://mainnet.infura.io/ws/v3/invalid")

    def test_set_network(self, test_init):
        self.client.set_network(network=self.network)
        assert self.client.is_connected()

    def test_get_network(self, test_init):
        assert self.client.get_network() == self.network

    def test_validate_address(self, test_init):
        assert self.client.validate_address(self.sushi_contract_address)
        assert self.client.validate_address(self.thor_token_address)

    def test_get_address(self, test_init):
        assert self.client.get_address() == self.test_address

    def test_set_phrase(self, test_init):
        assert self.client.set_phrase(self.mnemonic) == self.test_address
        assert self.client.is_connected()

    @pytest.mark.asyncio
    async def test_get_abi(self, test_init):
        if os.path.exists(self.prefix + self.thor_token_address):
            os.remove(self.prefix + self.thor_token_address)
        if os.path.exists(self.prefix + self.sushi_contract_address):
            os.remove(self.prefix + self.sushi_contract_address)
        assert str(await self.client.get_abi(self.sushi_contract_address)) == str(self.sushi_abi)
        assert str(await self.client.get_abi(self.thor_token_address)) == str(self.token_abi)
        self.client.ether_api = None
        assert str(await self.client.get_abi(self.sushi_contract_address)) == str(self.sushi_abi)
        assert str(await self.client.get_abi(self.thor_token_address)) == str(self.token_abi)

    @pytest.mark.asyncio
    async def test_get_contract(self, test_init):
        sushi_contract = await self.client.get_contract(self.sushi_contract_address, erc20=False)
        assert sushi_contract.functions.token0().call() == self.thor_token_address
        token_contract = await self.client.get_contract(self.thor_token_address, erc20=True)
        assert token_contract.functions.symbol().call() == 'RUNE'
        token_contract = await self.client.get_contract(self.thor_token_address, erc20=False)
        decimal_place = token_contract.functions.decimals().call()
        assert token_contract.functions.maxSupply().call()/10**decimal_place == 500000000.0

    @pytest.mark.asyncio
    async def test_get_balance(self, test_init):
        balance = await self.client.get_balance()
        balance_address = await self.client.get_balance(address=self.test_address)
        assert balance == balance_address
        # rune_balance = await self.client.get_balance(contract_address=self.thor_token_address)
        # rune_balance_address = await self.client.get_balance(address=self.test_address,
        #                                                contract_address=self.thor_token_address)
        # assert rune_balance == rune_balance_address

    # def test_get_fees(self, test_init):
    #     self.client.set_gas_strategy("fast")
    #     fast_fee = self.client.get_fees()
    #     self.client.set_gas_strategy("medium")
    #     medium_fee = self.client.get_fees()
    #     self.client.set_gas_strategy("slow")
    #     slow_fee = self.client.get_fees()
    #     assert fast_fee >= medium_fee >= slow_fee

    def test_get_transaction_data(self, test_init):
        data = self.client.get_transaction_data(self.test_hash)
        tx_hash = self.client.w3.toHex(data["hash"])
        from_addr = data["from"]
        value = data["value"]
        assert tx_hash == self.test_hash
        assert float(self.client.w3.fromWei(value, 'ether')) == 0.7

    def test_get_transaction_receipt(self, test_init):
        data = self.client.get_transaction_receipt(self.test_rune_transfer_hash)
        value = data['logs'][0]['data']
        amount = self.client.w3.toInt(hexstr=value)
        assert amount/10**18 == 1

    # def test_transfer_rune(self, test_init):
    #     self.client.set_gas_strategy("fast")
    #     dest_addr = self.client.w3.toChecksumAddress('0x5039c76445efcfa78d91b8974c100151634cbf2d')
    #     rune_balance = self.client.get_balance(contract_address=self.thor_token_address)
    #     assert rune_balance > 1
    #     receipt = self.client.transfer(dest_addr, 1, contract_address=self.thor_token_address)
    #     value = receipt['logs'][0]['data']
    #     amount = self.client.w3.toInt(hexstr=value)
    #     assert amount/10**18 == 1
    #
    # def test_transfer_ethereum(self, test_init):
    #     self.client.set_gas_strategy("fast")
    #     assert self.client.get_balance() > 0.0001
    #     tx_hash = self.client.transfer('0x81941E3DeEeA41b6309045ECbAFd919Db5aF6147', 0.0001)
    #     data = self.client.get_transaction_data(tx_hash)
    #     assert float(self.client.w3.fromWei(data["value"], 'ether')) == 0.0001

    @pytest.mark.asyncio
    async def test_read_contract(self, test_init):
        # assert await self.client.read_contract(self.thor_token_address, "balanceOf", self.client.get_address()) / 10**18\
        #        == await self.client.get_balance(contract_address=self.thor_token_address)
        assert await self.client.read_contract(self.sushi_contract_address, "factory", erc20=False) == "0xC0AEe478e3658e2610c5F7A4A2E1777cE9e4f2Ac"

    # def test_write_contract_token(self, test_init):
    #     old_balance = self.client.get_balance(contract_address=self.thor_token_address)
    #     func_to_call = "giveMeRUNE"
    #     self.client.set_gas_strategy("fast")
    #     tx_receipt = self.client.write_contract(self.thor_token_address, func_to_call, erc20=False)
    #     new_balance = self.client.get_balance(contract_address=self.thor_token_address)
    #     assert new_balance > old_balance

    # def test_write_contract_router(self):
    #     old_balance = self.client.get_balance(contract_address=self.thor_token_address)
    #     func_to_call = "deposit"
    #     asgard_address = self.client.w3.toChecksumAddress("0x5319f65449bf48cf979ba8e839e80415d2395284")
    #     amount = 1000*10**18
    #     memo = "switch:tthor1vecuqg5tejlxncykw6dfkj6hgkv49d59lc0z6j"
    #     x = self.client.get_contract(contract_address=self.thor_router_address, erc20=False)
    #     self.client.set_gas_strategy("fast")
    #     # <Function deposit(address,address,uint256,string)>
    #     tx_receipt = self.client.write_contract(self.thor_router_address, func_to_call,
    #                                             asgard_address, self.thor_token_address, amount,
    #                                             memo, erc20=False)
    #     new_balance = self.client.get_balance(contract_address=self.thor_token_address)
    #     assert new_balance < old_balance

    # def test_fast_write(self):
    #     func_to_call = "xxx"
    #     self.client.set_gas_strategy("fast")
    #     nonce = self.client.w3.eth.getTransactionCount(self.client.get_address())
    #     for i in range(10):
    #         tx_receipt = self.client.write_contract(self.thor_token_address, func_to_call, erc20=False, nonce=nonce+i)


import pytest
from xchainpy_ethereum.client import Client
from xchainpy_ethereum.models.asset import Asset
import json, os, time


ETH_RUNE = Asset("ETH", "RUNE-0xd601c6A3a36721320573885A8d8420746dA3d7A0")
ETH_ETH = Asset("ETH", "ETH")

class TestClient:
    prefix = "../xchainpy_ethereum/resources/ropsten/"
    mnemonic = open(prefix + "mnemonic", 'r').readline()
    router_abi = json.loads(open(prefix + "router_abi", 'r').readline())
    token_abi = json.loads(open(prefix + "token_abi", 'r').readline())
    network = open(prefix + "network", 'r').readline()
    eth_api = open("../xchainpy_ethereum/resources/ether_api", 'r').readline()
    thor_router_address = "0x9d496De78837f5a2bA64Cb40E62c19FBcB67f55a"

    test_address = "0x0dC1Ce70a8ddFA3F2070984C35010a285CF0530D"
    test_hash = "0x7aa171c7c024cbfdcf6e8097fbd8ec18bacf33581acdea0d3923c031a55b931e"
    test_rune_transfer_hash = "0xa23a19ec29fbc0edc245befe80bda6a29def5b3b4610bf20b670d9e90bfa8095"

    @pytest.fixture
    def test_init(self):
        self.client = Client(phrase=self.mnemonic, network=self.network, network_type="ropsten",
                             ether_api=self.eth_api)
        yield
        self.client.purge_client()

    def test_invalid_network_type(self):
        with pytest.raises(Exception) as err:
            assert Client(phrase=self.mnemonic, network=self.network, network_type="invalid")
        assert str(err.value) == "Network type has to be ropsten or mainnet"
        with pytest.raises(Exception) as err:
            assert Client(phrase=self.mnemonic, network=self.network, network_type="mainnet")
        assert str(err.value) == "invalid network type"

    def test_invalid_init_phrase(self):
        with pytest.raises(Exception) as err:
            assert Client(phrase="invalid", network=self.network)
        assert str(err.value) == "invalid phrase"

    def test_invalid_set_phrase(self, test_init):
        with pytest.raises(Exception) as err:
            assert self.client.set_phrase(phrase="invalid")
        assert str(err.value) == "invalid phrase"

    def test_is_connected(self, test_init):
        assert self.client.is_connected()

    def test_invalid_init_network(self):
        with pytest.raises(Exception) as err:
            assert Client(phrase=self.mnemonic, network="wss://ropsten.infura.io/ws/v3/invalid")

    def test_invalid_set_network(self, test_init):
        with pytest.raises(Exception) as err:
            assert self.client.set_network(network="wss://ropsten.infura.io/ws/v3/invalid")

    def test_set_network(self, test_init):
        self.client.set_network(network=self.network)
        assert self.client.is_connected()

    def test_get_network(self, test_init):
        assert self.client.get_network() == self.network

    def test_validate_address(self, test_init):
        assert self.client.validate_address(self.test_address)
        assert self.client.validate_address(self.thor_router_address)
        assert self.client.validate_address(ETH_RUNE.ticker)

    def test_get_address(self, test_init):
        assert self.client.get_address() == self.test_address

    def test_set_phrase(self, test_init):
        assert self.client.set_phrase(self.mnemonic) == self.test_address
        assert self.client.is_connected()

    @pytest.mark.asyncio
    async def test_get_abi(self, test_init):
        if os.path.exists(self.prefix + ETH_RUNE.ticker):
            os.remove(self.prefix + ETH_RUNE.ticker)
        if os.path.exists(self.prefix + self.thor_router_address):
            os.remove(self.prefix + self.thor_router_address)

        assert str(await self.client.get_abi(self.thor_router_address)) == str(self.router_abi)
        assert str(await self.client.get_abi(ETH_RUNE.ticker)) == str(self.token_abi)
        self.client.ether_api = None
        assert str(await self.client.get_abi(self.thor_router_address)) == str(self.router_abi)
        assert str(await self.client.get_abi(ETH_RUNE.ticker)) == str(self.token_abi)

    @pytest.mark.asyncio
    async def test_get_contract(self, test_init):
        router_contract = await self.client.get_contract(self.thor_router_address, erc20=False)
        assert router_contract.functions.RUNE().call() == ETH_RUNE.ticker
        token_contract = await self.client.get_contract(ETH_RUNE.ticker, erc20=True)
        assert token_contract.functions.symbol().call() == 'RUNE'
        token_contract = await self.client.get_contract(ETH_RUNE.ticker, erc20=False)
        decimal_place = token_contract.functions.decimals().call()
        assert token_contract.functions.maxSupply().call()/10**decimal_place == 500000000.0

    @pytest.mark.asyncio
    async def test_get_balance(self, test_init):
        balance = await self.client.get_balance()
        balance_address = await self.client.get_balance(address=self.test_address)
        assert balance == balance_address
        rune_balance = await self.client.get_balance(asset=ETH_RUNE)
        rune_balance_address = await self.client.get_balance(address=self.test_address,
                                                       asset=ETH_RUNE)
        assert rune_balance == rune_balance_address

    def test_get_fees(self, test_init):
        self.client.set_gas_strategy("fast")
        fast_fee = self.client.get_fees()
        self.client.set_gas_strategy("medium")
        medium_fee = self.client.get_fees()
        self.client.set_gas_strategy("slow")
        slow_fee = self.client.get_fees()
        assert fast_fee >= medium_fee >= slow_fee

    def test_get_transaction_data(self, test_init):
        data = self.client.get_transaction_data(self.test_hash)
        tx_hash = self.client.w3.toHex(data["hash"])
        from_addr = data["from"]
        value = data["value"]
        assert tx_hash == self.test_hash
        assert from_addr == self.test_address
        assert self.client.w3.fromWei(value, 'ether') == 3

    def test_get_transaction_receipt(self, test_init):
        data = self.client.get_transaction_receipt(self.test_rune_transfer_hash)
        value = data['logs'][0]['data']
        amount = self.client.w3.toInt(hexstr=value)
        assert amount/10**18 == 1

    @pytest.mark.asyncio
    async def test_transfer_rune(self, test_init):
        self.client.set_gas_strategy("fast")
        dest_addr = self.client.w3.toChecksumAddress('0x5039c76445efcfa78d91b8974c100151634cbf2d')
        rune_balance = await self.client.get_balance(asset=ETH_RUNE)
        assert rune_balance > 1
        receipt = await self.client.transfer(asset=ETH_RUNE, amount=1, recipient=dest_addr)
        value = receipt['logs'][0]['data']
        amount = self.client.w3.toInt(hexstr=value)
        assert amount/10**18 == 1

    @pytest.mark.asyncio
    async def test_transfer_ethereum(self, test_init):
        self.client.set_gas_strategy("fast")
        assert await self.client.get_balance() > 0.0001
        tx_hash = await self.client.transfer(asset=ETH_ETH, amount=0.0001, recipient='0x81941E3DeEeA41b6309045ECbAFd919Db5aF6147')
        data = self.client.get_transaction_data(tx_hash)
        assert float(self.client.w3.fromWei(data["value"], 'ether')) == 0.0001

    @pytest.mark.asyncio
    async def test_read_contract(self, test_init):
        assert await self.client.read_contract(ETH_RUNE.ticker, "balanceOf", self.client.get_address()) / 10**18 \
               == await self.client.get_balance(asset=ETH_RUNE)
        assert await self.client.read_contract(self.thor_router_address, "RUNE", erc20=False) == ETH_RUNE.ticker

    @pytest.mark.asyncio
    async def test_write_contract_token(self, test_init):
        old_balance = await self.client.get_balance(asset=ETH_RUNE)
        func_to_call = "giveMeRUNE"
        self.client.gas_price = self.client.w3.toWei(10, 'gwei')
        tx_receipt = await self.client.write_contract(ETH_RUNE.ticker, func_to_call, erc20=False)
        new_balance = await self.client.get_balance(asset=ETH_RUNE)
        time.sleep(1)
        assert new_balance > old_balance

    @pytest.mark.asyncio
    async def test_write_contract_router(self, test_init):
        old_balance = await self.client.get_balance(asset=ETH_RUNE)
        func_to_call = "deposit"
        asgard_address = self.client.w3.toChecksumAddress("0x8d6690d9068e0da603ada9c26d475f19b1d56c91")
        amount = 100*10**18
        memo = "switch:tthor1vecuqg5tejlxncykw6dfkj6hgkv49d59lc0z6j"
        router = await self.client.get_contract(contract_address=self.thor_router_address, erc20=False)
        self.client.gas_price = self.client.w3.toWei(10, 'gwei')
        # <Function deposit(address,address,uint256,string)>
        tx_receipt = await self.client.write_contract(self.thor_router_address, func_to_call,
                                                asgard_address, ETH_RUNE.ticker, amount,
                                                memo, erc20=False)
        new_balance = await self.client.get_balance(asset=ETH_RUNE)
        time.sleep(1)
        assert new_balance < old_balance

    # @pytest.mark.asyncio
    # async def test_thorchain_eth_rune(self, test_init):
    #     old_balance = await self.client.get_balance()
    #     func_to_call = "approve"


    # def test_fast_write(self):
    #     func_to_call = "xxx"
    #     self.client.set_gas_strategy("fast")
    #     nonce = self.client.w3.eth.getTransactionCount(self.client.get_address())
    #     for i in range(10):
    #         tx_receipt = self.client.write_contract(ETH_RUNE.ticker, func_to_call, erc20=False, nonce=nonce+i)


from xchainpy_util.asset import Asset as template
from xchainpy_util.chain import is_chain


class Asset(template):
    def __init__(self, chain, symbol, ticker = ''):
        """
        :param chain: chain type
        :type chain: str
        :param symbol: symbol name
        :type symbol: str
        :param ticker: is the symbol or the contract address of the token
        :type ticker: str
        """

        if is_chain(chain):
            self._chain = chain
        else:
            raise Exception('the chain is invalid')
        self._symbol = symbol
        if not ticker:
            if '-' in symbol:
                self._ticker = symbol[symbol.index('-')+1:]
            else:
                self._ticker = symbol
        else:
            self._ticker = ticker


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='xchainpy_ethereum',
    packages=find_packages(),
    package_data={
        "xchainpy_ethereum": ["resources/*"],
    },
    include_package_data=True,
    version='0.2.1',
    license='MIT',
    description='Custom Ethereum client and utilities used by XChainPY clients',
    author='THORChain',
    author_email='',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/xchainjs/xchainpy-lib/tree/main/xchainpy/xchainpy_ethereum',
    keywords=["ETH", "Ethereum", "XChainpy_ethereum","THORChain", "web3"],
    install_requires=required,
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python'
    ],
)


import pytest
from xchainpy_ethereum.models.asset import Asset

class TestAsset:

    eth_asset = Asset('ETH', 'ETH-0x0000000000000000000000000000000000000000', '0x0000000000000000000000000000000000000000')

    def test_create_asset(self):
        asset = Asset('ETH', 'USDT-0XA3910454BF2CB59B8B3A401589A3BACC5CA42306')
        assert asset.chain == 'ETH'
        assert asset.symbol == 'USDT-0XA3910454BF2CB59B8B3A401589A3BACC5CA42306'
        assert asset.ticker == '0XA3910454BF2CB59B8B3A401589A3BACC5CA42306'

    def test_invalid_chain(self):
        with pytest.raises(Exception) as err:
            Asset('invalid chain', 'ETH', 'ETH')
        assert str(err.value) == "the chain is invalid"

    def test_asset_to_string(self):
        asset = Asset('ETH', 'USDT-0XA3910454BF2CB59B8B3A401589A3BACC5CA42306', '0XA3910454BF2CB59B8B3A401589A3BACC5CA42306')
        assert str(asset) == 'ETH.USDT-0XA3910454BF2CB59B8B3A401589A3BACC5CA42306'

    def test_asset_equal(self):
        asset = Asset('ETH', 'USDT-0XA3910454BF2CB59B8B3A401589A3BACC5CA42306', '0XA3910454BF2CB59B8B3A401589A3BACC5CA42306')
        asset2 = Asset('ETH', 'USDT-0XA3910454BF2CB59B8B3A401589A3BACC5CA42306', '0XA3910454BF2CB59B8B3A401589A3BACC5CA42306')
        assert asset == asset2